---
session_store_driver:
  Directory:
    path: {{ config_dir }}/sessions/
ownership_voucher_store_driver:
  Directory:
    path: {{ test_dir }}/ownership_vouchers/
trusted_device_keys_path: {{ keys_path }}/device_ca_cert.pem
owner_private_key_path: {{ keys_path }}/owner_key.der
owner_public_key_path: {{ keys_path }}/owner_cert.pem
owner_addresses:
- transport: HTTP
  port: 8079
  addresses:
    - dns_name: localhost
- transport: HTTP
  port: {{ owner_port }}
  addresses:
    - dns_name: localhost
report_to_rendezvous_endpoint_enabled: true
bind: {{ bind }}
service_info_api_url: "http://localhost:{{ serviceinfo_api_server_port }}/device_info"
service_info_api_authentication:
  BearerToken:
    token: TestAuthToken
