---
session_store_driver:
  Directory:
    path: {{ config_dir }}/sessions/
ownership_voucher_store_driver:
  Directory:
    path: {{ test_dir }}/ownership_vouchers/
bind: {{ bind }}
rendezvous_info:
- dns: localhost
  device_port: 8082
  owner_port: 8082
  protocol: http
- dns: localhost
  delay: 30
  device_port: {{ rendezvous_port }}
  owner_port: {{ rendezvous_port }}
  protocol: http
protocols:
  diun:
    key_path: {{ keys_path }}/diun_key.der
    cert_path: {{ keys_path }}/diun_cert.pem
    key_type: SECP256R1
    mfg_string_type: SerialNumber
    allowed_key_storage_types:
    - {{ diun_key_type }}
manufacturing:
  manufacturer_cert_path: {{ keys_path }}/manufacturer_cert.pem
  manufacturer_private_key: {{ keys_path }}/manufacturer_key.der
  owner_cert_path: {{ keys_path }}/owner_cert.pem
  device_cert_ca_private_key: {{ keys_path }}/device_ca_key.der
  device_cert_ca_chain: {{ keys_path }}/device_ca_cert.pem
